/*
    Copyright (C) 2003 <ryu@gpul.org>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <osgCal/CoreModel>

#include <osgDB/Registry>
#include <osgDB/Input>
#include <osgDB/Output>
#include <osgDB/FileUtils>

using namespace osg;
using namespace osgDB;
using namespace osgCal;

// forward declare functions to use later.
bool CoreModel_readLocalData(Object& obj, Input& fr);
bool CoreModel_writeLocalData(const Object& obj, Output& fw);

bool CoreModel_matchModeStr(const char* str,int& mode);
const char* CoreModel_getModeStr(int value);

RegisterDotOsgWrapperProxy g_CoreModelProxy
(
    new osgCal::CoreModel,
    "CoreModel",
    "Object CoreModel",
    &CoreModel_readLocalData,
    &CoreModel_writeLocalData
);

bool CoreModel_readLocalData(Object& obj, Input& fr)
{
    bool iteratorAdvanced = false;

    CoreModel& model = static_cast<CoreModel&>(obj);

    if (fr[0].matchWord("skeleton") && fr[1].isString())
    {
	std::string path = osgDB::findDataFile(fr[1].getStr());
	if (path.empty()) {
	    std::cerr << "osgCal ERROR: Can't find the skeleton file '" << fr[1].getStr() << "'!" << std::endl;
	} else {
	    model.loadSkeleton(path);
	}
        fr+=2;
        iteratorAdvanced = true;
    }

    if (fr[0].matchWord("animation") && fr[1].isString())
    {
	std::string path = osgDB::findDataFile(fr[1].getStr());
	if (path.empty()) {
	    std::cerr << "osgCal ERROR: Can't find the animation file '" << fr[1].getStr() << "'!" << std::endl;
	} else {
	    model.loadAnimation(path);
	}
        fr+=2;
        iteratorAdvanced = true;
    }

    if (fr[0].matchWord("mesh") && fr[1].isString())
    {
	std::string path = osgDB::findDataFile(fr[1].getStr());
	if (path.empty()) {
	    std::cerr << "osgCal ERROR: Can't find the mesh file '" << fr[1].getStr() << "'!" << std::endl;
	} else {
	    model.loadMesh(path);
	}
        fr+=2;
        iteratorAdvanced = true;
    }

    if (fr[0].matchWord("material") && fr[1].isString())
    {
	std::string path = osgDB::findDataFile(fr[1].getStr());
	if (path.empty()) {
	    std::cerr << "osgCal ERROR: Can't find the material file '" << fr[1].getStr() << "'!" << std::endl;
	} else {
	    model.loadMaterial(path);
	}
        fr+=2;
        iteratorAdvanced = true;
    }
    
    return iteratorAdvanced;
}

bool CoreModel_writeLocalData(const Object& obj, Output& fw)
{

#if 0
    const CoreModel& model = static_cast<const CoreModel&>(obj);
    std::vector<std::string> items;

    fw.indent() << "skeleton " << fw.wrapString(fw.getFileNameForOutput( model.getSkeletonFileName() ))<< std::endl;
    items=model.getAnimationFileNames();
    unsigned i; // For visual sux.. ehhm studio :P
    for (i=0;i<items.size(); i++) {
	fw.indent() << "animation " << fw.wrapString(fw.getFileNameForOutput( items[i] ))<< std::endl;
    }
    items=model.getMeshFileNames();
    for (i=0;i<items.size(); i++) { //###################
	fw.indent() << "mesh " << fw.wrapString(fw.getFileNameForOutput( items[i] ))<< std::endl;
    }
    items=model.getMaterialFileNames();
    for (i=0;i<items.size(); i++) {
	fw.indent() << "material " << fw.wrapString(fw.getFileNameForOutput( items[i] ))<< std::endl;
    }

    return true;
#else
    return true;
#endif
}
